
from syned.beamline.optical_element import OpticalElement
from syned.beamline.shape import BoundaryShape

class Absorber(OpticalElement):
    """
    Base class for optical element category "absorbers" (filters, slits, etc.)

    Constructor.

    Parameters
    ----------
    name : str
        The name of the optical element.
    boundary_shape : instance of BoundaryShape, optional
        if None, it is initialized to BoundaryShape().

    """
    def __init__(self, name="Undefined", boundary_shape=None):
        if boundary_shape is None: boundary_shape = BoundaryShape()
        OpticalElement.__init__(self, name=name, boundary_shape=boundary_shape)



from syned.syned_object import SynedObject

class MagneticStructure(SynedObject):
    """
    Base clase for magnetic structures (from where BM, Wiggler and Undulator will heritate)
    """
    def __init__(self):
        super().__init__()



"""
Base classes for optical elements with shape(s).

   * OpticalElementsWithSurfaceShape (e.g. an elliptical mirror).
   * OpticalElementsWithMultipleShapes (e.g. a lens that has two surfaces).
"""
from syned.beamline.optical_element import OpticalElement
from syned.beamline.shape import SurfaceShape

class OpticalElementsWithMultipleShapes(OpticalElement):
    """
    Constructor.

    Parameters
    ----------
    name : str
        The optical elemnt name.
    surface_shapes : list
        List with surface shapes (instances of SurfaceShape)
    boundary_shape : instance of BoundaryShape, optional
        The element shape. The default=None means no shape associated to the optical element.

    """

    def __init__(self, name="",
                 surface_shapes=None,
                 boundary_shape=None):
        super().__init__(name, boundary_shape)
        if surface_shapes is None: surface_shapes = [SurfaceShape()]
        if not isinstance(surface_shapes, list): raise ValueError("surface_shapes must of type 'list'")
        self._surface_shapes = surface_shapes

    def get_surface_shape(self, index):
        """
        Gets the surface shape.

        Parameters
        ----------
        index : int
            The index of the requested surface shape in the list.

        Returns
        -------
        instance of SurfaceShape
            The requested surface shape.

        """
        try: return self._surface_shapes[index]
        except: raise Exception("only " + str(len(self._surface_shapes)) + " shapes in OpticalElementsWithMultipleShapes")

    def set_surface_shape(self, index, surface_shape=None):
        """
        Sets a surface shape.

        Parameters
        ----------
        index : int
            The index of the surface shape to be set in the list.
        surface_shape : instances of SurfaceShape, optional
            The surface shape to be set (If None, initialize to SurfaceShape())

        """
        if surface_shape is None: surface_shape = SurfaceShape()
        try: self._surface_shapes[index] = surface_shape
        except: raise Exception("only " + str(len(self._surface_shapes)) + " shapes in OpticalElementsWithMultipleShapes")

class OpticalElementsWithSurfaceShape(OpticalElementsWithMultipleShapes):
    """
    Constructor.

    Parameters
    ----------
    name : str
        The optical elemnt name.
    surface_shape : instances of SurfaceShape
        The surface shape.
    boundary_shape : instance of BoundaryShape, optional
        The element shape. The default=None means no shape associated to the optical element.
    """
    def __init__(self, name, surface_shape=None, boundary_shape=None):
        super(OpticalElementsWithSurfaceShape, self).__init__(name=name,
                                                              boundary_shape=boundary_shape,
                                                              surface_shapes=[surface_shape])

    # these definitions are necessary to define self._surface_shape which is needed to
    # build the object instance from json files.
    @property
    def _surface_shape(self):
        return self.get_surface_shape()

    @_surface_shape.setter
    def _surface_shape(self, value):
        self.set_surface_shape(value)

    def get_surface_shape(self):
        """
        Gets the surface shape.

        Returns
        -------
        instance of SurfaceShape
            The requested surface shape.

        """
        return super(OpticalElementsWithSurfaceShape, self).get_surface_shape(index=0)

    def set_surface_shape(self, surface_shape=None):
        """
        Sets a surface shape.

        Parameters
        ----------
        surface_shape : instances of SurfaceShape, optional
            The surface shape to be set (If None, initialize to SurfaceShape())

        """
        super(OpticalElementsWithSurfaceShape, self).set_surface_shape(index=0, surface_shape=surface_shape)


if __name__=="__main__":
    from syned.beamline.shape import Cylinder, EllipticalCylinder

    oe = OpticalElementsWithSurfaceShape(name="TEST", surface_shape=Cylinder())
    print(oe.get_surface_shape())
    oe.set_surface_shape(surface_shape=EllipticalCylinder())
    print(oe.get_surface_shape())




from syned.storage_ring.electron_beam import ElectronBeam
from syned.storage_ring.magnetic_structures.undulator import Undulator
from syned.storage_ring.light_source import LightSource
from syned.beamline.shape import *

from syned.beamline.optical_elements.ideal_elements.screen import Screen
from syned.beamline.optical_elements.ideal_elements.ideal_lens import IdealLens
from syned.beamline.optical_elements.absorbers.filter import Filter
from syned.beamline.optical_elements.absorbers.slit import Slit
from syned.beamline.optical_elements.absorbers.beam_stopper import BeamStopper

from syned.beamline.optical_elements.mirrors.mirror import Mirror
from syned.beamline.optical_elements.crystals.crystal import Crystal
from syned.beamline.optical_elements.gratings.grating import Grating

from syned.beamline.beamline import BeamlineElement, Beamline
from syned.beamline.element_coordinates import ElementCoordinates

from syned.util.json_tools import load_from_json_file

#
# example of setting a beamline in SYNED.
#
if __name__ == "__main__":


    print("==================== LightSource: ==================")

    src1 = ElectronBeam.initialize_as_pencil_beam(energy_in_GeV=6.0,current=0.2)

    src2 = Undulator()
    src2.set_value_from_key_name("K_horizontal",33)
    # just to be sure...
    assert (33 == src2.get_value_from_key_name("K_horizontal"))

    src = LightSource("test",src1,src2)

    # check file o/i for test
    src.to_json("tmp.json") # write to file
    tmp = load_from_json_file("tmp.json") # read from file
    print("returned class: ", type(tmp))
    print(src.to_dictionary())
    print(tmp.to_dictionary())
    assert(src.to_dictionary() == tmp.to_dictionary())


    print("==================== Optical elements: ==================")

    #
    # ideal elements
    #

    screen1 = Screen("screen1")
    lens1 = IdealLens("lens1",3.0)

    #
    # absorbers
    #
    filter1 = Filter("filter1","H2O",3.0e-6)

    slit1 = Slit(name="slit1",boundary_shape=Rectangle(-0.5e-3,0.5e-3,-2e-3,2e-3))

    slit2 = Slit(name="slit2")
    slit2.set_rectangle(width=3e-4,height=5e-4)
    slit2.set_circle(radius=3e-4)

    stopper1 = BeamStopper(name="stopper1",boundary_shape=Rectangle(-0.5e-3,0.5e-3,-2e-3,2e-3))

    stopper2 = BeamStopper(name="stopper2")

    stopper2.set_rectangle(width=3e-4,height=5e-4)
    stopper2.set_circle(radius=3e-4)

    #
    # elements with shape: mirror, gratings, crystals
    #

    mirror1 = Mirror(name="mirror1")

    crystal1 = Crystal(name="crystal1")

    grating1 = Grating(name="grating1")


    print("==================== BeamLine: ==================")

    beamline1 = Beamline()

    beamline1.set_light_source(src)

    list_oe = [screen1, lens1, filter1, slit1, slit2, stopper1, stopper2, mirror1, crystal1, grating1]
    for i, optical_element in enumerate(list_oe):
        coordinates=ElementCoordinates(p=1.1*i, q=1.2*i)
        be = BeamlineElement(optical_element=optical_element, coordinates=coordinates)
        beamline1.append_beamline_element(be)

    print(beamline1.info())

    # check file o/i for test
    beamline1.to_json("tmp_beamline1.json")
    tmp = load_from_json_file("tmp_beamline1.json")
    print("returned class: ",type(tmp))
    print(beamline1.to_dictionary())
    print(tmp.to_dictionary())
    assert(beamline1.to_dictionary() == tmp.to_dictionary())



"""
Position coordinates of a beamline component.
"""

from syned.syned_object import SynedObject

class ElementCoordinates(SynedObject):
    """

    """
    def __init__(self, p = 0.0, q = 0.0, angle_radial=0.0, angle_azimuthal=0.0, angle_radial_out=None):
        """

        Parameters
        ----------
        p : float, optional
            distance from previous element in m.
        q : float, optional
            distance to next element in m.
        angle_radial : float, optional
            Radial inclination angle in rads.
        angle_azimuthal : float, optional
            Azimuthal inclination angle in rads.
        angle_radial_out : float, optional
            The radial angle in rads in the output direction (default=None, the same as angle_radial).

        """
        self._p               = p
        self._q               = q
        self._angle_radial    = angle_radial
        self._angle_azimuthal = angle_azimuthal
        self._angle_radial_out = angle_radial_out

        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("p",                "distance from previous continuation plane", "m"    ),
                    ("q",                "distance to next continuation plane",       "m"    ),
                    ("angle_radial",     "incident angle [to normal]",                "rad"  ),
                    ("angle_radial_out", "output angle [to normal]",                  "rad"),
                    ("angle_azimuthal",  "rotation along beam axis",                  "rad"  ),
                ])

    def p(self):
        """
        Returns the distance from previous element.

        Returns
        -------
        float

        """
        return self._p

    def q(self):
        """
        Returns the distance to next element.

        Returns
        -------
        float

        """
        return self._q

    def angle_radial(self):
        """
        Returns the radial angle.

        Returns
        -------
        float

        """
        return self._angle_radial

    def angle_radial_out(self):
        """
        Returns the radial angle in the output direction.

        Returns
        -------
        float

        """
        if self._angle_radial_out is None:
            return self.angle_radial()
        else:
            return self._angle_radial_out

    def angle_azimuthal(self):
        """
        Returns the azimuthal angle.

        Returns
        -------
        float

        """
        return self._angle_azimuthal

    def set_positions(self, p=0.0, q=0.0, angle_radial=0.0, angle_radial_out=None, angle_azimuthal=0.0):
        """
        Sets the coordinates.

        Parameters
        ----------
        p : float, optional
            distance from previous element in m.
        q : float, optional
            distance to next element in m.
        angle_radial : float, optional
            Radial inclination angle in rads.
        angle_azimuthal : float, optional
            Azimuthal inclination angle in rads.
        angle_radial_out : float, optional
            The radial angle in rads in the output direction (default=None, the same as angle_radial).

        """
        self._p = p
        self._q = q
        self._angle_radial = angle_radial
        self._angle_radial_out = angle_radial_out
        self._angle_azimuthal = angle_azimuthal

    def get_positions(self):
        """
        Gets the coordinates.

        Returns
        -------
        tuple
            (p, q, angle_radial, angle_radial_out, angle_azimuthal)

        """
        return self.p(), \
            self.q(), \
            self.angle_radial(), \
            self.angle_radial_out(), \
            self.angle_azimuthal()

    def set_p_and_q(self, p=0.0, q=0.0):
        """
        Set the distances p and q.

        Parameters
        ----------
        p : float, optional
            distance from previous element in m.
        q : float, optional
            distance to next element in m.

        """
        self._p = p
        self._q = q

    def get_p_and_q(self):
        """
        Gets p and q.

        Returns
        -------
        tuple
            (p,q).

        """
        return self.p(), self.q()

    def set_angles(self, angle_radial=0.0, angle_radial_out=None, angle_azimuthal=0.0):
        """
        Sets the angles.

        Parameters
        ----------
        angle_radial : float, optional
            Radial inclination angle in rads.
        angle_azimuthal : float, optional
            Azimuthal inclination angle in rads.
        angle_radial_out : float, optional
            The radial angle in rads in the output direction (default=None, the same as angle_radial).

        Returns
        -------

        """
        self._angle_radial = angle_radial
        self._angle_radial_out = angle_radial_out
        self._angle_azimuthal = angle_azimuthal

    def get_angles(self):
        """
        Get the angles.

        Returns
        -------
        tuple
            (angle_radial, angle_radial_out, angle_azimuthal)

        """
        return self.angle_radial(), self.angle_radial_out(), self.angle_azimuthal()



"""
Functions to read/write syned objects in json files.

Notes
-----
The elements that are not explicitely imported in the code that follows, but they MUST be imported
to define the objects at run time.
To import more elements, yoy can use the exec_command keyword.

"""
from syned.storage_ring.electron_beam import ElectronBeam
from syned.storage_ring.magnetic_structures.undulator import Undulator
from syned.storage_ring.magnetic_structures.wiggler import Wiggler
from syned.storage_ring.magnetic_structures.bending_magnet import BendingMagnet
from syned.storage_ring.magnetic_structure import MagneticStructure
from syned.beamline.optical_elements.ideal_elements.screen import Screen
from syned.beamline.optical_elements.ideal_elements.ideal_lens import IdealLens
from syned.beamline.optical_elements.absorbers.filter import Filter
from syned.beamline.optical_elements.absorbers.slit import Slit
from syned.beamline.optical_elements.absorbers.beam_stopper import BeamStopper
from syned.beamline.optical_elements.absorbers.holed_filter import HoledFilter
from syned.beamline.optical_elements.mirrors.mirror import Mirror
from syned.beamline.optical_elements.crystals.crystal import Crystal
from syned.beamline.optical_elements.gratings.grating import Grating
from syned.beamline.optical_elements.gratings.grating import GratingVLS
from syned.beamline.optical_elements.gratings.grating import GratingBlaze
from syned.beamline.optical_elements.gratings.grating import GratingLamellar

from syned.beamline.shape import BoundaryShape
from syned.beamline.shape import Rectangle, Ellipse, Circle
from syned.beamline.shape import MultiplePatch
from syned.beamline.shape import DoubleRectangle, DoubleEllipse, DoubleCircle

from syned.beamline.shape import SurfaceShape
from syned.beamline.shape import Conic, Plane, Sphere, SphericalCylinder
from syned.beamline.shape import Ellipsoid, EllipticalCylinder
from syned.beamline.shape import Paraboloid, ParabolicCylinder
from syned.beamline.shape import Hyperboloid, HyperbolicCylinder
from syned.beamline.shape import Toroid

from syned.storage_ring.light_source import LightSource
from syned.storage_ring.empty_light_source import EmptyLightSource

from syned.beamline.beamline import Beamline
from syned.beamline.beamline_element import BeamlineElement
from syned.beamline.element_coordinates import ElementCoordinates

from collections import OrderedDict

try:
    import json_tricks as json # to save numpy arrays
except:
    import json
from urllib.request import urlopen

def load_from_json_file(file_name, exec_commands=None):
    """
    Function to load a syned object from a json file.

    Parameters
    ----------
    file_name : str
        The file name.
    exec_commands : str
        The commands (typically import...) to be executed before accesing the file.

    Returns
    -------
    instance of SynedObject

    """
    f = open(file_name)
    text = f.read()
    f.close()
    return load_from_json_text(text, exec_commands=exec_commands)

def load_from_json_url(file_url, exec_commands=None):
    """
    Function to load a syned object from a remote json file.

    Parameters
    ----------
    file_url : str
        The URL with the file name.
    exec_commands : str
        The commands (typically import...) to be executed before accesing the file.

    Returns
    -------
    instance of SynedObject

    """
    u = urlopen(file_url)
    ur = u.read()
    url = ur.decode(encoding='UTF-8')
    return load_from_json_text(url, exec_commands=exec_commands)

def load_from_json_text(text, exec_commands=None):
    """
    Function to load a syned object from a json txt.

    Parameters
    ----------
    text : str
        The text with the corresponding info.
    exec_commands : str
        The commands (typically import...) to be executed before accesing the file.

    Returns
    -------
    instance of SynedObject

    """
    return load_from_json_dictionary_recurrent(json.loads(text), exec_commands=exec_commands)

def load_from_json_dictionary_recurrent(jsn, verbose=False, exec_commands=None):
    """
    Function to convert a dictionary (got from json file) into a syned object.

    Parameters
    ----------
    jsn : dict
        The dictionary with json file information.
    verbose : boolean, optional
        Define or not verbose output.
    exec_commands : str
        The commands (typically import...) to be executed before accesing the file.

    Returns
    -------
    instance of SynedObject

    """
    if isinstance(exec_commands, list):
        for command in exec_commands:
            if verbose: print(">>>>",command)
            exec(command)
    elif isinstance(exec_commands, str):
        if verbose: print(">>>>", exec_commands)
        exec(exec_commands)

    if verbose: print(jsn.keys())
    if "CLASS_NAME" in jsn.keys():
        if verbose: print("FOUND CLASS NAME: ",jsn["CLASS_NAME"])
        if verbose: print(">>>>eval: ", jsn["CLASS_NAME"])
        try:
            tmp1 = eval(jsn["CLASS_NAME"]+"()")
        except:
            raise RuntimeError("Error evaluating: "+jsn["CLASS_NAME"]+"() ** you could use the exec_command keyword to import it at run time **")


        if tmp1.keys() is not None:

            for key in tmp1.keys():
                if verbose: print(">>>>processing",key ,type(jsn[key]))
                if isinstance(jsn[key],dict):
                    if verbose: print(">>>>>>>>dictionary found, starting recurrency",key ,type(jsn[key]))
                    tmp2 = load_from_json_dictionary_recurrent(jsn[key],exec_commands=exec_commands)
                    if verbose: print(">>>>2",key,type(tmp2))
                    tmp1.set_value_from_key_name(key,tmp2)
                elif isinstance(jsn[key], list):
                    if verbose: print(">>>>>>>>LIST found, starting recurrency",key ,type(jsn[key]))
                    out_list_of_objects = []
                    for element in jsn[key]:
                        if isinstance(element, dict):
                            if verbose: print(">>>>>>>>LIST found, starting recurrency",key ,type(element))
                            tmp3 = load_from_json_dictionary_recurrent(element, exec_commands=exec_commands)
                            if verbose: print(">>>>3",type(tmp3))
                            out_list_of_objects.append(tmp3)
                        else:
                            print("***** Failed to load", element)
                    if verbose: print("kkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkk",out_list_of_objects)
                    tmp1.set_value_from_key_name(key,out_list_of_objects)
                        # tmp1.set_value_from_key_name(key,tmp2)
                else:
                    if verbose: print(">>>>>>> setting value for key: ",key," to: ",repr(jsn[key]))
                    tmp1.set_value_from_key_name(key,jsn[key])

        return tmp1


if __name__ == "__main__":

    file_url = "http://ftp.esrf.fr/pub/scisoft/syned/lightsources/ESRF_ID01_EBS_PPU35_22.json"
    syned_obj = load_from_json_url(file_url)
    print(syned_obj.info())


    #
    # file_url = "/home/manuel/Oasys/ALSU-IR-BM.json"
    # syned_obj = load_from_json_file(file_url)
    # print(syned_obj.info())

    # file_url = "/home/manuel/Oasys/bl.json"
    # syned_obj = load_from_json_file(file_url)
    # print(syned_obj.info())



"""
Base class for LighSource, which contains:
    - a name
    - an electron beam
    - a magnetic structure

"""
from syned.syned_object import SynedObject


class EmptyLightSource(SynedObject):
    """
    Defines an empty light source. This is for creating "movable" beamlines (using Beamline()). These are beamlines
    that do not have a particular light source.

    Parameters
    ----------
    name : str, optional
        The name of the (empty) light source.

    """
    def __init__(self, name="Empty"):
        self._name               = name
        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("name",              "Name",""),
            ] )


    def get_name(self):
        """
        Returns the name of the light source.

        Returns
        -------
        str

        """
        return self._name




import numpy
from syned.beamline.optical_elements.ideal_elements.ideal_element import IdealElement

class IdealFZP(IdealElement):
    """
    Defines an ideal Fresnel Zone Plate.

    Constructor.

    Parameters
    ----------
    name : str, optional
        The name of the optical element.
    focusing_direction : int
        0=None, 1=x (sagittal), 2=z (meridional), 3=2D focusing.
    focal : float
        The focal length in meters.
    nominal_wavelength : float
        The nominal wavelength in m for where the focal length is defined.
    diameter : float
        The FZP diameter in m.

    """
    def __init__(self,
                 name="Undefined",
                 focusing_direction=3,  # 0=None, 1=x (sagittal), 2=z (meridional), 3=2D focusing.
                 focal=1.0,                 #  focal distance (m)
                 nominal_wavelength=1e-10,  # nominal wavelength in m
                 # r0=10.0e-6,                #  inner zone radius (m)
                 diameter=0.001,            # FZP diameter in m
                 ):
        IdealElement.__init__(self, name=name)
        self._focusing_direction = focusing_direction
        self._focal = focal
        self._nominal_wavelength = nominal_wavelength
        self._diameter = diameter

        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("name"              , "Name"                , ""),
                    ("boundary_shape"    , ""                    , ""),
                    ("focusing_direction", "Focusing direction: 0=None, 1=1D along X, 2=1D along Z, 3=2D", ""),
                    ("focal"             , "Focal length"        , "m"),
                    ("nominal_wavelength", "Nominal wavelength"  , "m"),
                    ("diameter"          , "FZP diameter"        , "m"),
            ] )

    def focusing_direction(self):
        """
        Returns the focusing direction.

        Returns
        -------
        int
            0=None,
            1=1D along X,
            2=1D along Z,
            3=2D.
        """
        return self._focusing_direction

    def focal(self):
        """
        Returns the focal length.

        Returns
        -------
        float

        """
        return self._focal

    def nominal_wavelength(self):
        """
        Returns the nominal wavelength.

        Returns
        -------
        float

        """
        return self._nominal_wavelength

    def diameter(self):
        """
        Returns the FZP diameter.

        Returns
        -------
        float

        """
        return self._diameter

    #
    # calculated. The exact expression is rn = sqrt( n lambda f + (n lambda / 2)**2 )
    #
    def rn(self):
        return 0.5 * self.diameter()

    def r0(self):
        """
        Returns the innermost radius (approximated calculation r0=sqrt(wavelength * focal)).

        Returns
        -------
        float

        """
        return numpy.sqrt(self.nominal_wavelength() * self.focal())

    def r0_exact(self):
        """
        Returns the innermost radius (exact calculation r0=sqrt(wavelength * focal + (wavelength/2)**2)).

        Returns
        -------
        float

        """
        return numpy.sqrt(self.nominal_wavelength() * self.focal() + (0.5 * self.nominal_wavelength())**2)

    def n_vs_r(self, r): # approximated if f >> diameter; eq 855 in Michette
        """
        Returns the zone number for a given distance r (approximated calculation).

        Returns
        -------
        float

        """
        return (r ** 2 / self.nominal_wavelength() / self.focal())

    def n_exact_vs_r(self, r): # Exact, solving n from rn = sqrt( n lambda f + (n lambda / 2)**2 )
        """
        Returns the zone number for a given distance r
        (exact calculation solving n from rn = sqrt( n lambda f + (n lambda / 2)**2).

        Returns
        -------
        float

        """
        nn = -2 * self.focal() + 2 * numpy.sqrt(self.focal()**2 + r**2)
        return nn / self.nominal_wavelength()

    def n(self):
        """
        Returns the zone number for a outermost zone (approximated calculation).

        Returns
        -------
        float

        """
        return self.n_vs_r( self.rn() )

    def n_exact(self):
        """
        Returns the zone number for a outermost zone (exact calculation using n_exact_vs_r() ).

        Returns
        -------
        float

        """
        return self.n_exact_vs_r(self.rn())

if __name__ == "__main__":
    fzp = IdealFZP(
        name = "Undefined",
        focusing_direction = 3,  # 0=None, 1=x (sagittal), 2=z (meridional), 3=2D focusing.
        focal = 1.0,  # focal distance (m)
        nominal_wavelength = 1e-10,  # nominal wavelength in m
        diameter = 0.001,  # FZP diameter in m
    )

    print(fzp.info())
    print("r0, r0_exact: ", fzp.r0(), fzp.r0_exact())
    print("rn: ", fzp.rn())
    print("n, n_exact: ", fzp.n(), fzp.n_exact())






from syned.beamline.shape import BoundaryShape
from syned.beamline.shape import Rectangle, Ellipse

from syned.beamline.optical_elements.absorbers.absorber import Absorber

class BeamStopper(Absorber):
    """
    Beam-stopper or obstruction.

    Constructor.

    Note that:
                                   Slit     BeamStopper    Filter      HoledFilter
        beam pass at center         Yes     No             Yes         No
        apply attenuation           No      No             Yes         Yes

    Parameters
    ----------
    name : str
        The name of the optical element.
    boundary_shape : instance of BoundaryShape, optional
        The geometry of the slit aperture. if None, it is initialized to BoundaryShape().

    """
    def __init__(self, name="Undefined", boundary_shape=None):
        if boundary_shape is None: boundary_shape = BoundaryShape()
        Absorber.__init__(self, name=name, boundary_shape=boundary_shape)

    def set_rectangle(self,width=3e-3,height=4e-3):
        """
        Sets the stopper as a rectangle.

        Parameters
        ----------
        width : float, optional
            The rectangle width.
        length : float, optional
            The rectangle length.
        center_x : float, optional
            The center coordinate X.
        center_y : float, optional
            The center coordinate Y.

        """
        self._boundary_shape=Rectangle(-0.5*width,0.5*width,-0.5*height,0.5*height)

    def set_circle(self,radius=3e-3):
        """
        Sets the stopper as a circle.

        Parameters
        ----------
        radius : float
            The radius of the circle.
        center_x : float
            The x coordinate of the center of the circle.
        center_y   : float
            The y coordinate of the center of the circle.

        """
        self._boundary_shape=Ellipse(-0.5*radius,0.5*radius,-0.5*radius,0.5*radius)

    #TODO: add set_ellipse (like in slit).


from syned.beamline.optical_elements.refractors.lens import Lens

class CRL(Lens):
    """
    Defines a Compound refractive lens (CRL). It is composed by a number (n_lens) of identical lenses.

    Constructor.

    Parameters
    ----------
    name : str, optional
        The name of the optical element.
    n_lens : int, optional
        The number of (identical) lenses.
    surface_shape1 : instance of SurfaceShape, optional
        The geometry of the lens surface 1. if None, it is initialized to SurfaceShape().
    surface_shape2 : instance of SurfaceShape, optional
        The geometry of the lens surface 2. if None, it is initialized to SurfaceShape().
    boundary_shape : instance of BoundaryShape, optional
        The geometry of the slit aperture. if None, it is initialized to BoundaryShape().
    material : str
        A string defining the material within the two surfaces.
    thickness : float
        The distance between the two surfaces at the center of the lens in m.
    piling_thickness : float, optional
        The piling distance in m, or spatial periodicity in the stack. In other words, the distance from the
        lens1-surface1 to the lens2-surface1.

    """
    def __init__(self,
                 name="Undefined",
                 n_lens=1,
                 surface_shape1=None,
                 surface_shape2=None,
                 boundary_shape=None,
                 material="",
                 thickness=0.0,
                 piling_thickness=0.0):
        super().__init__(name=name,
                         surface_shape1=surface_shape1,
                         surface_shape2=surface_shape2,
                         boundary_shape=boundary_shape,
                         material=material,
                         thickness=thickness)
        self._n_lens           = n_lens
        self._piling_thickness = piling_thickness

        # support text contaning name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("name",                "Name" ,                                "" ),
                    ("n_lens",              "N Lens" ,                              "" ),
                    ("surface_shapes",      "Surface shapes",                       "" ),
                    ("boundary_shape",      "Boundary shape",                       "" ),
                    ("material",            "Material (element, compound or name)", "" ),
                    ("thickness",           "Thickness",                            "m"),
                    ("piling_thickness",    "Piling Thickness",                     "m")
            ] )

    def get_n_lens(self):
        """
        Returns the number of lenses.

        Returns
        -------
        int

        """
        return self._n_lens

    def get_piling_thickness(self):
        """
        Returns the piling thickness.

        Returns
        -------
        float

        """
        return self._piling_thickness
